"""Command line interface for YamlTable.

See https://docs.python.org/3/using/cmdline.html#cmdoption-m for why module is
named __main__.py.
"""


import pathlib
import pprint
from typing import List, Optional, Tuple

import typer

import yamltable
from yamltable.typing import ExitCode, FileArg, Row, Schema, StatusColor


app = typer.Typer(
    help=(
        "Utility for working with YAML files organized similar to a relational"
        " database table."
    )
)


@app.command(name="index")
def index_(index: int, file_path: pathlib.Path = FileArg) -> None:
    """Get row at INDEX in FILE_PATH."""
    rows, _ = load_data(file_path)

    try:
        row = rows[index]
    except IndexError:
        typer.secho(
            f"Error: Index {index} is out of bounds.",
            fg=StatusColor.ERROR.value,
            err=True,
        )
        raise typer.Exit(code=ExitCode.ERROR.value)
    else:
        typer.secho(pprint.pformat(row, indent=2))


@app.command(name="list")
def list_(key: str, file_path: pathlib.Path = FileArg) -> None:
    """List all dictionary KEY values in FILE_PATH."""
    rows, _ = load_data(file_path)

    for idx, row in enumerate(rows):
        try:
            typer.secho(row[key])
        except KeyError:
            typer.secho(
                f"Error: Row {idx} does not have key {key}.",
                fg=StatusColor.ERROR.value,
                err=True,
            )
            raise typer.Exit(code=ExitCode.ERROR.value)


def load_data(file_path: pathlib.Path) -> Tuple[List[Row], Optional[Schema]]:
    """Attempt to load data from YAML file.

    Args:
        file_path: YAML file path

    Returns:
        YAML row data, YAML schema
    """
    try:
        return yamltable.read(file_path)
    except (FileNotFoundError, TypeError) as xcpt:
        typer.secho(f"Error: {xcpt}", fg=StatusColor.ERROR.value, err=True)
        raise typer.Exit(code=ExitCode.ERROR.value)


@app.command()
def search(key: str, value: str, file_path: pathlib.Path = FileArg) -> None:
    """Search dictionaries in FILE_PATH with matching KEY and VALUE pairs."""
    rows, _ = load_data(file_path)
    matches = yamltable.search(key, value, rows)

    if matches:
        for match in yamltable.search(key, value, rows):
            typer.secho(pprint.pformat(match, indent=2))
    else:
        typer.secho(
            f"No rows found with (key={key}, value={value}) pair.",
            fg=StatusColor.EMPTY.value,
        )


@app.command()
def sort(key: str, file_path: pathlib.Path = FileArg) -> None:
    """Sort dictionaries in FILE_PATH by KEY values."""
    rows, schema = load_data(file_path)

    try:
        sorted_rows = yamltable.sort(key, rows)
    except TypeError as xcpt:
        typer.secho(f"Error: {xcpt}", fg=StatusColor.ERROR.value, err=True)
        raise typer.Exit(code=ExitCode.ERROR.value)
    else:
        yamltable.write(file_path, sorted_rows, schema)


@app.command()
def validate(file_path: pathlib.Path = FileArg) -> None:
    """Check that every dictionary in FILE_PATH has conforms to its schema."""
    rows, schema = load_data(file_path)
    valid, row, msg = yamltable.validate(rows, schema)

    if valid:
        typer.secho(
            "YAML file rows conform to its schema.",
            fg=StatusColor.SUCCESS.value,
        )
    elif row == -1:
        typer.secho(
            f"Invalid schema: {msg}", fg=StatusColor.ERROR.value, err=True
        )
        raise typer.Exit(code=ExitCode.ERROR.value)
    else:
        typer.secho(
            f"Invalid row {row}: {msg}", fg=StatusColor.ERROR.value, err=True
        )
        raise typer.Exit(code=ExitCode.INVALID.value)


@app.command()
def version() -> None:
    """Display application version."""
    typer.secho(f"yamltable {yamltable.__version__}")


if __name__ == "__main__":  # pragma: no cover
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['yamltable']

package_data = \
{'': ['*']}

install_requires = \
['jsonschema>=3.2.0,<4.0.0',
 'pyrsistent>=0.14.11,<0.15.0',
 'pyyaml>=5.2,<6.0',
 'typer>=0.0.11,<0.0.12']

entry_points = \
{'console_scripts': ['yamltable = yamltable.__main__:app']}

setup_kwargs = {
    'name': 'yamltable',
    'version': '0.0.10',
    'description': 'Command line utility for list organized YAML files.',
    'long_description': '# YamlTable\n\n![](https://img.shields.io/pypi/v/yamltable)\n![](https://img.shields.io/pypi/pyversions/yamltable.svg)\n![](https://github.com/wolfgangwazzlestrauss/yamltable/workflows/build/badge.svg)\n![](https://codecov.io/gh/wolfgangwazzlestrauss/yamltable/branch/master/graph/badge.svg)\n![](https://img.shields.io/badge/code%20style-black-000000.svg)\n![](https://img.shields.io/github/repo-size/wolfgangwazzlestrauss/yamltable)\n![](https://img.shields.io/github/license/wolfgangwazzlestrauss/yamltable)\n\n---\n\n**Documentation**: https://wolfgangwazzlestrauss.github.io/yamltable\n\n**Source Code**: https://github.com/wolfgangwazzlestrauss/yamltable\n\n---\n\nYamlTable is a Python command line utility for working with YAML files organized\nsimilar to a relational database table. It supports YAML files organized as a\nlist of dictionaries, which share key names and value types. YamlTable provides\ncommands for listing, searching, sorting, etc. data from the supported files.\n\n## Supported YAML File Organizations\n\nYamlTable works with YAML files organized as a list of dictionaries with similar\nkey names and value types.\n\n```yaml\n- name: awscli\n  description: Amazon Web Services command line client\n  website: https://aws.amazon.com/\n- name: glances\n  description: operating system monitoring interface\n  website: https://github.com/nicolargo/glances\n```\n\nThe JSON schema support is included for YAML files organized as:\n\n```yaml\nschema:\n  $schema: http://json-schema.org/draft-07/schema#\n  description: pipx package metadata schema\n  type: object\n  properties:\n    name:\n      type: string\n      pattern: "^[\\\\w-]+$"\n    description:\n      type: string\n    website:\n      type: string\n  required:\n    - name\n    - description\n    - website\n  additionalProperties: false\nrows:\n  - name: awscli\n    description: Amazon Web Services command line client\n    website: https://aws.amazon.com/\n  - name: glances\n    description: operating system monitoring interface\n    website: https://github.com/nicolargo/glances\n```\n\n## Getting Started\n\n### Installation\n\nYamlTable can be installed for Python 3.6+ with `pipx`.\n\n```bash\npipx install yamltable\n```\n\nTo reuse its library functions install with `pip`.\n\n```bash\npip install --user yamltable\n```\n\n### Commands\n\nYamlTable provides the following commands for working with YAML files:\n\n- `list`: list dictionary key values\n- `search`: search dictionaries by key and value\n- `sort`: sort dictionaries by key and value\n- `validate`: validate that dictionaries conform to the given JSON schema\n\n## Contributing\n\nSince YamlTable is in an early development phase, it is not currently open to\ncontributors.\n\n## License\n\nLicensed under the [MIT](license.txt) license.\n',
    'author': 'Macklan Weinstein',
    'author_email': 'wolfgangwazzlestrauss@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/wolfgangwazzlestrauss/yamltable',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""Command line interface for sorting YAML files."""


import pathlib
from typing import Any, IO, Iterable, List, Optional, Sequence, Tuple, Union

import jsonschema
from jsonschema import exceptions
import yaml

from yamltable.typing import Row, Schema


__author__ = "Macklan Weinstein"
__version__ = "0.0.10"


def dependencies(unsorted: Sequence[Row], depends: str, name: str) -> List[Row]:
    """Sort rows based on the dependencies found in key.

    Args:
        unsorted: Rows to sort.
        depends: Dictionary key whose values are other dictionary dependencies.
        name: Foreign key for dependencies.

    Raises:
        ValueError: If rows contain circular requirements.

    Returns:
        Rows sorted by requirements.
    """
    sorted_: List[Row] = []
    while unsorted:
        # Find rows containing only dependencies in sorted_.
        names = [pkg[name] for pkg in sorted_]
        rows = [
            (idx, pkg)
            for idx, pkg in enumerate(unsorted)
            if set(pkg[depends]).issubset(names)
        ]

        if rows:
            idxs, pkgs = zip(*rows)
            # Append found rows to sorted_ and remove them from unsorted.
            sorted_ += pkgs
            unsorted = [
                elem for idx, elem in enumerate(unsorted) if idx not in idxs
            ]
        else:
            # If no rows are found then there exist circular dependencies.
            raise ValueError("encountered circular dependencies.")

    return sorted_


def read(
    stream: Union[IO[str], pathlib.Path, str]
) -> Tuple[List[Row], Optional[Schema]]:
    """Read data from YAML file.

    Args:
        stream: YAML text, text I/O stream, or file path.

    Raises:
        FileNotFoundError: If unable to find file path.
        TypeError: If file is not organized as a list.

    Returns:
        YAML data.
    """
    if isinstance(stream, pathlib.Path):
        stream = stream.read_text()

    try:
        data = yaml.safe_load(stream)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as xcpt:
        raise TypeError(f"invalid YAML file: {xcpt}")

    if isinstance(data, list):
        return data, None
    elif isinstance(data, dict):
        try:
            return data["rows"], data["schema"]
        except KeyError:
            raise TypeError(
                "YAML file does not have a schema and rows organization"
            )
    else:
        raise TypeError("YAML file is not organized in a tabular format")


def search(key: str, val: Any, rows: Iterable[Row]) -> List[Row]:
    """Search dictionaries by key and value.

    Args:
        key: Search key.
        val: Key comparison value.
        rows: Dictionaries to search.

    Returns:
        Matching dictionaries.

    Examples:
        >>> rows = [{"foo": 1, "bar": 2}, {"foo": 5, "bar": 6}]
        >>> search("bar", 6, rows)
        [{'foo': 5, 'bar': 6}]
    """
    return [row for row in rows if key in row and row[key] == val]


def sort(key: str, rows: Iterable[Row]) -> List[Row]:
    """Sort dictionaries based on value for supplied key name.

    Args:
        key: Search key.
        rows: Dictionaries to sort.

    Returns:
        List of sorted dictionaries.

    Examples:
        >>> rows = [{"foo": 5, "bar": 2}, {"foo": 3, "bar": 6}]
        >>> sort("foo", rows)
        [{'foo': 3, 'bar': 6}, {'foo': 5, 'bar': 2}]
        >>> sort("bar", rows)
        [{'foo': 5, 'bar': 2}, {'foo': 3, 'bar': 6}]
    """
    return sorted(rows, key=lambda row: row[key])


def validate(
    rows: Iterable[Row], schema: Optional[Schema]
) -> Tuple[bool, int, str]:
    """Check that each row satisfies the schema.

    Args:
        rows: Dictionaries to validate.
        schema: JSON schema for validation.

    Returns:
        Whether all rows are valid, invalid row index or -1,
            invalid error message.
    """
    try:
        jsonschema.Draft7Validator.check_schema(schema)
    except (exceptions.SchemaError, exceptions.UnknownType) as xcpt:
        error_msg = str(xcpt)
        return False, -1, error_msg

    validator = jsonschema.Draft7Validator(schema)
    for idx, row in enumerate(rows):
        try:
            validator.validate(row)
        except exceptions.ValidationError as xcpt:
            error_msg = str(xcpt)
            return False, idx, error_msg

    return True, -1, ""


def write(
    file_path: pathlib.Path,
    rows: List[Row],
    schema: Optional[Schema] = None,
    sort_keys: bool = False,
) -> None:
    """Write data to YAML file.

    Args:
        file_path: YAML file path.
        rows: List of dictionaries to write.
        schema: JSON schema dictionary.
        sort_keys: Whether to sort row keys.
    """
    if schema is None:
        with open(file_path, "w") as handle:
            yaml.dump(rows, handle, sort_keys=sort_keys)
    else:
        data = {"schema": schema, "rows": rows}
        with open(file_path, "w") as handle:
            yaml.dump(data, handle, sort_keys=sort_keys)



"""Library types for YamlTable."""


import enum
from typing import Any, Dict

import typer

Row = Dict[str, Any]
Schema = Dict[str, Any]


class ExitCode(enum.Enum):
    """Exit code statuses."""

    SUCCESS = 0
    INVALID = 1
    ERROR = 2


FileArg = typer.Argument(
    ..., dir_okay=False, exists=True, file_okay=True, resolve_path=True
)


class StatusColor(enum.Enum):
    """Colors for message types."""

    EMPTY = typer.colors.YELLOW
    ERROR = typer.colors.RED
    SUCCESS = typer.colors.BRIGHT_GREEN
